#!/bin/sh
# the next line restarts using tclsh \
    exec tclsh "$0" ${1+"$@"}

# leg20180329
#
# Based on D. Bohdans' JSON value extraction benchmark page in the
# Tcler's Wiki: http://wiki.tcl.tk/48500
#
# Run benchmarks for JSON parsers.
#
# This file defines test functions for all parser implementations and
# how to extract a specific value from the JSON data for each of them.
#
# It also defines the test data files for each test to be run.
#
# And finally it contains the benchmark machinery.
#
# To run a benchmark, you must specify a Tcl file as argument, which
# sets the variable 'testSeries' to a list of 'testname' 'times' 'unit'
# triples.
#
# * The 'testname' must be specified in this file and refers to a
#   specific JSON data set.
# * 'times' is the number of times the test shall be run.
# * 'unit' can be: µs, ms or s, and defines, in which of these units
#   the test results shall be shown.
#
# A second variable 'testImps' must be set to a list of JSON parser
# implementations as defined in this file.  The implementations are
# specified by their proc name and run in the order they are listed.
#
#

package require fileutil
puts {Package versions:}
puts "Tcl           --  [package require Tcl]"

#####
# Define all JSON parsing functions here.
#
# Each function must have the form
#
#   proc procName {key data}
#
# It will be given a JSON string in data and a key to extract from
# there in key.  The proc must return the respective value from the
# JSON data.
#
# You must also provide two functions for setup and teardown of the package.
#  proc procName_setup {} .. loads the package and does any initialization, if needed
#  proc procName_teardown {} .. frees ressources, if needed
#
#  If anything of this is not needed define the proc empty.

### jq
proc jq_setup {} {
    puts "jq            --  should print version here."
}
proc jq_teardown {} {}
proc jq {key data} {return [exec jq -r $key << $data]}

### duktape
proc duktape_setup {} {
    puts "duktape       --  [package require duktape]"
    package require duktape::oo
    set ::duk [::duktape::oo::Duktape new]
}
proc duktape_teardown {} {$::duk destroy}
proc duktape {key data} {
    set j [::duktape::oo::JSON new $::duk $data]
    set result [$j get {*}$key]
    $j destroy
    return $result
}

### jimhttp
proc jimhttp-json_setup {} {
    source jimhttp/json.tcl
    puts "jimhttp json  --  $::json::version"
}
proc jimhttp-json_teardown {} {}
proc jimhttp-json {key data} {
    return [dict get [::json::parse $data] {*}$key]
}

### rl_json
proc rl_json_setup {} {
    puts "rl_json       --  [package require rl_json]"
}
proc rl_json_teardown {} {}
proc rl_json {key data} {
    return [::rl_json::json get $data {*}key]
}

### Tcllib
proc tcllib-json_setup {} {
    puts "Tcllib json   --  [package require json]"
    json::LoadAccelerator tcl
    json::SwitchTo tcl
}
proc tcllib-json_teardown {} {}
proc tcllib-json {key data} {
    set parsed [::json::json2dict $data]
    return [eval $key]
}

### SQLite
proc sqlite3-json1_setup {} {
    puts "sqlite3       --  [package require sqlite3]"
    sqlite3 ::sq3 :memory:
    ::sq3 enable_load_extension 1
}
proc sqlite3-json1_teardown {} {::sq3 close}
proc sqlite3-json1 {key data} {
    return [lindex [::sq3 eval {
        select json_extract($data, $key)
    }] 0]
}

### yajltcl
proc yajltcl_setup {} {
    puts "yajltcl       --  [package require yajltcl]"
}
proc yajltcl_teardown {} {}
proc yajltcl {key data} {
    set parsed [::yajl::json2dict $data]
    return [eval $key]
}

### Alternative JSON
proc jsonDecodeTcl_setup {} {
    source jsonDecode.tcl
    puts "jsonDecode    --  Tcler's wiki"
}
proc jsonDecodeTcl_teardown {} {}
proc jsonDecodeTcl {key data} {
    set parsed [jsonDecode $data]
    return [eval $key]
}

##### TON
proc load_ton {} {
    if {![namespace exists ::ton]} {
	source ton.tcl
	puts "ton           --  $::ton::version"
    }
}

### TON: rl, validating, 2list
proc ton2list_setup {} {load_ton}
proc ton2list_teardown {} {}
proc ton2list {key data} {
    set parsed [namespace eval ton::2list [ton::json2ton $data]]
    return [ton::2list::get $parsed {*}$key]
}

### TON: rl, validating, 2dict
proc ton2dict_setup {} {load_ton}
proc ton2dict_teardown {} {}
proc ton2dict {key data} {
    set parsed [namespace eval ton::2dict [ton::json2ton $data]]
    return [eval $key]
 }

### TON: rl, validating, a2dict
proc tona2dict_setup {} {load_ton}
proc tona2dict_teardown {} {}
proc tona2dict {key data} {
    set parsed [namespace eval ton::a2dict [ton::json2ton $data]]
    return [dict get $parsed {*}$key]
}

##### TON: left to right parser
proc load_tonlr {} {
    load_ton
    if {![namespace exists ::ton::lr]} {
	source ton.lr.tcl
	puts "ton l to r    --  $::ton::lr::version"
    }
}
### TON: lr, validating, 2list
proc tonlr2list_setup {} {load_tonlr}
proc tonlr2list_teardown {} {}
proc tonlr2list {key data} {
    set parsed [namespace eval ton::2list [ton::lr::json2ton $data]]
    return [ton::2list::get $parsed {*}$key]
}
### TON: lr, validating, 2dict
proc tonlr2dict_setup {} {load_tonlr}
proc tonlr2dict_teardown {} {}
proc tonlr2dict {key data} {
    set parsed [namespace eval ton::2dict [ton::lr::json2ton $data]]
    return [eval $key]
 }
### TON: lr, validating, a2dict
proc tonlra2dict_setup {} {load_tonlr}
proc tonlra2dict_teardown {} {}
proc tonlra2dict {key data} {
    set parsed [namespace eval ton::a2dict [ton::lr::json2ton $data]]
    return [dict get $parsed {*}$key]
}

##### TON: non validating parser
proc load_tonnv {} {
    load_ton
    if {![namespace exists ::ton::nv]} {
	source ton.nv.tcl
	puts "ton l to r    --  $::ton::nv::version"
    }
}
### TON: nv, validating, 2list
proc tonnv2list_setup {} {load_tonnv}
proc tonnv2list_teardown {} {}
proc tonnv2list {key data} {
    set parsed [namespace eval ton::2list [ton::nv::json2ton $data]]
    return [ton::2list::get $parsed {*}$key]
}
### TON: nv, validating, 2dict
proc tonnv2dict_setup {} {load_tonnv}
proc tonnv2dict_teardown {} {}
proc tonnv2dict {key data} {
    set parsed [namespace eval ton::2dict [ton::nv::json2ton $data]]
    return [eval $key]
 }
### TON: nv, validating, a2dict
proc tonnva2dict_setup {} {load_tonnv}
proc tonnva2dict_teardown {} {}
proc tonnva2dict {key data} {
    set parsed [namespace eval ton::a2dict [ton::nv::json2ton $data]]
    return [dict get $parsed {*}$key]
}

##### TON: non recursive parser
proc load_tonnr {} {
    load_ton
    if {![namespace exists ::ton::nr]} {
	source ton.nr.tcl
	puts "ton nr        --  $::ton::nr::version"
    }
}
### TON: nr, validating, 2list
proc tonnr2list_setup {} {load_tonnr}
proc tonnr2list_teardown {} {}
proc tonnr2list {key data} {
    set parsed [namespace eval ton::2list [ton::nr::json2ton $data]]
    return [ton::2list::get $parsed {*}$key]
}
### TON: nr, validating, 2dict
proc tonnr2dict_setup {} {load_tonnr}
proc tonnr2dict_teardown {} {}
proc tonnr2dict {key data} {
    set parsed [namespace eval ton::2dict [ton::nr::json2ton $data]]
    return [eval $key]
 }
### TON: nr, validating, a2dict
proc tonnra2dict_setup {} {load_tonnr}
proc tonnra2dict_teardown {} {}
proc tonnra2dict {key data} {
    set parsed [namespace eval ton::a2dict [ton::nr::json2ton $data]]
    return [dict get $parsed {*}$key]
}


### json2dict, https://wiki.tcl.tk/13419
proc json2dict JSONtext {
    string range \
	[string trim \
	     [string trimleft \
		  [string map {
		      \t {} \n {} \r {}
		      , { } : { } \[ \{ \] \}
		  } $JSONtext] {\uFEFF}
	     ]
	] 1 end-1
}
proc j2d_setup {} {
    puts "json2dict     --  Tcler's wiki"
}
proc j2d_teardown {} {}
proc j2d {key data} {
    set parsed [json2dict $data]
    return [eval $key]
}


#####
# Configuration of the tests and the test functions
#
# Each test is specified by a tag, the file where the JSON data is
# read from and the expected result.  The name attribute is used to
# identify the test in the benchmark output.
#
# This data is stored in the global 'ds' (dataset) array
foreach {tag name filename result} {
    rfc_o {RFC 7158 example object} test.json 38793
    mock_100 {Mockaroo with 100 items} mock_100.json Male
    mock_1000 {Mockaroo with 1000 items} mock_1000.json kkruscheveu@sfgate.com
    json-generator {json-generator.com repeat(10000, 100)} \
	generated.json "Rosales Osborne"
    mtg {Magic The Game Card Data} AllSets.json \
	{Children claim no two feathers are exactly the same color,\
	     then eagerly gather them for proof.}
} {
    set ds($tag,name) $name
    set ds($tag,filename) $filename
    set ds($tag,value) $result
}

# Now we need to specify for each JSON parser function, how to
# retrieve the test result for each test data.  We do this in the
# global 'imp' array, which uses the proc name as key, and also holds
# a friendly name for the parser function, a setup and a teardown function.

# The following is the method required to get a specific json path
# from the dict returned by duktape, jimhttp, rl_json and ton2list
set dictStyleSetup {
    rfc_o  {Image IDs 3}
    mock_100 {43 gender}
    mock_1000 {534 email}
    json-generator {1302 friends 1 name}
    mtg {INV cards 68 flavor}
}
# The following is the method required to get a specific json path
# from the dict returned by tcllib, yajltcl and ton2dict
set tcllibStyleSetup {
    rfc_o {lindex [dict get $parsed Image IDs] 3}
    mock_100 {dict get [lindex $parsed 43] gender}
    mock_1000 {dict get [lindex $parsed 534] email}
    json-generator {dict get \
			[lindex [dict get [lindex $parsed 1302] \
				     friends] \
			     1] \
			name}
    mtg {dict get [lindex [dict get $parsed INV cards] 68] flavor}
}

# jq, sqlite3-json1, jsonDecode and tona2dict have their own retrieval
# method.

foreach {procName name extractValue} \
    [subst -nocommand {
 	tcllib-json {Tcllib JSON} {$tcllibStyleSetup}
	yajltcl yajl-tcl {$tcllibStyleSetup}
	ton2dict ton2dict {$tcllibStyleSetup}
	tonlr2dict {ton2dict lr} {$tcllibStyleSetup}
	tonnv2dict {ton2dict nv} {$tcllibStyleSetup}
	tonnr2dict {ton2dict nr} {$tcllibStyleSetup}
	j2d json2dict {$tcllibStyleSetup}

	tcl-duktape duktape {$dictStyleSetup}
	jimhttp-json {jimhttp JSON} {$dictStyleSetup}
	rl_json rl_json {$dictStyleSetup}
	ton2list ton2list {$dictStyleSetup}
	tona2dict tona2dict {$dictStyleSetup}
	tonlr2list {ton2list lr} {$dictStyleSetup}
	tonlra2dict {tona2dict lr} {$dictStyleSetup}
	tonnv2list {ton2list nv} {$dictStyleSetup}
	tonnva2dict {tona2dict nv} {$dictStyleSetup}
	tonnr2list {ton2list nr} {$dictStyleSetup}
	tonnra2dict {tona2dict nr} {$dictStyleSetup}

	jq jq {
	    rfc_o {.Image.IDs[3]}
	    mock_100 {[43].gender}
	    mock_1000 {[534].email}
	    json-generator {[1302].friends.[1].name}
	    mtg {.INV.cards[68].flavor}    
	}
	sqlite3-json1 {SQLite JSON1} {
	    rfc_o {$.Image.IDs[3]}
	    mock_100 {$[43].gender}
	    mock_1000 {$[534].email}
	    json-generator {$[1302].friends[1].name}
	    mtg {$.INV.cards[68].flavor}
	}
	jsonDecodeTcl jsonDecode {
	    rfc_o {dict get \
		      [lindex \
			   [dict get \$parsed object \
				Image object \
				IDs array] \
			   3] number}
	    mock_100 {dict get \
			  [lindex [dict get \$parsed array] 43] \
			  object gender string}

	    mock_1000 {dict get \
			   [lindex [dict get \$parsed array] 534] \
			   object email string}
	    json-generator \
		{dict get \
		     [lindex [dict get [lindex [dict get \$parsed array] 1302] \
				  object friends array] 1] object name string}
	    mtg {dict get \
		     [lindex \
			  [dict get \
			       \$parsed object INV object cards array] \
			  68] object \
		     flavor string}
	}
    }] {
	set imp($procName,name) $name
	set imp($procName,setup) ${procName}_setup
	set imp($procName,teardown) ${procName}_teardown
	foreach {test key} $extractValue {
	    set imp($procName,$test,key) $key
	}
    }
 

#####
# Utility functions
#

proc setup {testImps} {
    foreach imp $testImps {uplevel #0 ${imp}_setup}
    puts {}
}
proc teardown {testImps} {
    foreach imp $testImps {uplevel #0 ${imp}_teardown}
}

proc benchmark {test times procName data result} {
    global imp
    return [time {
        set actualResult [$procName $imp($procName,$test,key) $data]
        if {$actualResult ne $result} {error "bad result: \"$actualResult\""}
    } $times]
}

proc main {testSeries testImps} {
    global imp
    global ds
   
    foreach {test times scale} $testSeries {
	puts "Running test '$ds($test,name)' with $times iterations for each library"
	foreach procName $testImps {
	    set data [::fileutil::cat $ds($test,filename)]
	    puts -nonewline [format {%-15s} $imp($procName,name)]
	    flush stdout
	    # Hack to allow json2dict to compete
	    if {$test eq "mtg" && $procName eq "j2d"} {
		set ds(mtg,value) {Children claim no two feathers are exactly the same color  then eagerly gather them for proof.}
	    } else {
		set ds(mtg,value) {Children claim no two feathers are exactly the same color, then eagerly gather them for proof.}
	    }
	    set t [lindex [benchmark $test $times \
			       $procName $data $ds($test,value)] 0]
	    switch -exact $scale {
		µs {set unit μs/run}
		ms {
		    set unit ms/run
		    set t [expr {$t/1000}]
		}
		s {
		    set unit s/run
		    set t [expr {$t/1000000}]
		}
	    }
	    puts [format {%13.2f %s} $t $unit]
	}
	puts {}
    }
}

if {[llength $argv]!=1 && ![file readable [lindex $argv 0]]} {
    error "Please specify a file with the list of tests and implementations to run."
}
    
source [lindex $argv 0]


setup $testImps
main $testSeries $testImps
teardown $testImps

